@@ -51,20 +51,13 @@ def getAuthTokenFromRequest(request, auth_strategy):
 """
 Helper function to fetch the public key for the given app id from Passage
 """
-def fetchPublicKey(app_id):
+def fetchApp(app_id):
     # unauthenticated request to get the public key
     r = requests.get(BASE_URL + app_id)
 
     # check response code
     if r.status_code != 200:
-        raise PassageError("Could not fetch public key for app id " + app_id)
+        raise PassageError("Could not fetch app information for app id " + app_id)
 
-    try:
-        public_key = r.json()["app"]["rsa_public_key"]
-        origin = r.json()["app"]["auth_origin"]
-        keyBytes = b64decode(public_key)
-        pubKey = load_pem_public_key(keyBytes, default_backend())
-        return pubKey, origin
-    except Exception as e:
-        raise PassageError("Could not fetch public key for app id " + app_id)
+    return r.json()["app"]
 
