@@ -9,11 +9,11 @@ else:
     from typing_extensions import TypeDict, Union
 
 from requests.sessions import Request
-from passageidentity.helper import extractToken, getAuthTokenFromRequest, fetchPublicKey
+from passageidentity.helper import fetchApp, getAuthTokenFromRequest
 from passageidentity.errors import PassageError
 from enum import Enum
 
-PUBKEY_CACHE = {}
+AUTH_CACHE = {}
 BASE_URL = "https://api.passage.id/v1/apps/"
 
 class UserStatus(Enum):
@@ -54,6 +54,16 @@ class Passage():
         recent_events: Union[None, list]
         webauthn: bool
         webauthn_devices: Union[None, list]
+        user_metadata: dict
+    class PassageDeviceType(TypedDict):
+        created_at: str
+        updated_at: str
+        id: str
+        cred_id: str
+        friendly_name: str
+        usage_count: int
+    
+        
 
     """
     When a Passage object is created, fetch the public key from the cache or make an API request to get it
@@ -67,14 +77,36 @@ class Passage():
             raise PassageError("Passage App ID must be provided")
 
         # if the pubkey exists in the cache, use that to avoid making requests
-        if app_id in PUBKEY_CACHE.keys():
-            self.passage_pubkey: str = PUBKEY_CACHE[app_id]["public_key"]
-            self.auth_origin: str = PUBKEY_CACHE[app_id]["auth_origin"]
+        if app_id in AUTH_CACHE.keys():
+            self.jwks: str = AUTH_CACHE[app_id]["jwks"]
+            self.auth_origin: str = AUTH_CACHE[app_id]["auth_origin"]
         else:
-            self.passage_pubkey, self.auth_origin = fetchPublicKey(app_id)
-            PUBKEY_CACHE[app_id] = {"public_key": self.passage_pubkey, "auth_origin": self.auth_origin}
+            self.__refreshAuthCache()
     
 
+    """
+    Fetch JWKs for the app
+    """
+    def __fetchJWKS(self):
+        r = requests.get(f"https://auth.passage.id/v1/apps/{self.app_id}/.well-known/jwks.json")
+
+        if r.status_code != 200:
+            raise PassageError("Could not fetch JWKs for app id " + self.app_id)
+
+        jwks = r.json()["keys"]
+
+        # translate the JWKS into map for O(1) access
+        jwkItems = {}
+        for jwk in jwks:
+            jwkItems[jwk["kid"]] = jwk
+
+        return jwkItems
+
+    def __refreshAuthCache(self):
+        self.auth_origin = fetchApp(self.app_id)["auth_origin"]
+        self.jwks = self.__fetchJWKS()
+        AUTH_CACHE[self.app_id] = {"jwks": self.jwks, "auth_origin": self.auth_origin}
+
     """
     Authenticate a Flask or Django request that uses Passage for authentication.
     This function will verify the JWT and return the user ID for the authenticated user, or throw
@@ -88,8 +120,8 @@ class Passage():
 
         # load and parse the JWT
         try:
-            claims = jwt.decode(token, self.passage_pubkey, audience=self.auth_origin, algorithms=["RS256"])
-            return claims["sub"]
+            userID = self.authenticateJWT(token)
+            return userID
         except Exception as e:
             raise PassageError("JWT is not valid: " + str(e))
 
@@ -102,7 +134,18 @@ class Passage():
     def authenticateJWT(self, token:str) -> Union[str, PassageError]:
         # load and parse the JWT
         try:
-            claims = jwt.decode(token, self.passage_pubkey, audience=self.auth_origin, algorithms=["RS256"])
+            kid = jwt.get_unverified_header(token)["kid"]
+            jwk = AUTH_CACHE[self.app_id]["jwks"][kid]
+
+            # if the JWK can't be found, they might need to udpate the JWKS for this Passage intance
+            # re-fetch the JWKS and try again
+            if not jwk:
+                self.__refreshAuthCache
+                kid = jwt.get_unverified_header(token)["kid"]
+                jwk = AUTH_CACHE[self.app_id]["jwks"][kid]
+
+            public_key = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
+            claims = jwt.decode(token, public_key, audience=self.auth_origin, algorithms=["RS256"])
             return claims["sub"]
         except Exception as e:
             raise PassageError("JWT is not valid: " + str(e)) 
@@ -160,6 +203,52 @@ class Passage():
             return PassageUser(user_id, r.json()["user"])
         except Exception as e:
             raise PassageError("Could not fetch user data")
+    
+    """
+    Use Passage API to list user devices, look up by user ID
+    """ 
+    def listUserDevices(self, user_id: str) -> Union[list, PassageError]:
+        if self.passage_apikey == "":
+            raise PassageError("No Passage API key provided.")
+
+        header = {"Authorization": "Bearer " + self.passage_apikey}
+        try:
+            url = BASE_URL + self.app_id + "/users/" + user_id + "/devices"
+            r = requests.get(url, headers=header)
+
+            if r.status_code != 200:
+                # get error message
+                message = r.json()["status"]
+                raise PassageError(f"{message}")
+            device_list = list()
+            devices = r.json()["devices"]
+            if devices != None:
+                for d in devices:
+                    device_list.append(PassageDevice(d))
+            return device_list 
+        except Exception as e:
+            raise PassageError("Could not list user's devices: {e}")
+
+    """
+    Use Passage API to list user devices, look up by user ID
+    """ 
+    def revokeUserDevice(self, user_id: str, device_id: str) -> Union[bool, PassageError]:
+        if self.passage_apikey == "":
+            raise PassageError("No Passage API key provided.")
+
+        header = {"Authorization": "Bearer " + self.passage_apikey}
+        try:
+            url = BASE_URL + self.app_id + "/users/" + user_id + "/devices/" + device_id
+            r = requests.delete(url, headers=header)
+
+            if r.status_code != 200:
+                # get error message
+                message = r.json()["status"]
+                raise PassageError(f"{message}")
+            return True 
+        except Exception as e:
+            raise PassageError("Could not revoke user device: {e}")
+
 
 
     """
@@ -207,11 +296,12 @@ class Passage():
     """
     Update Passage User's Attributes
     """
-    class UpdateUserAttributes(TypedDict):
-        phone: str
+    class UserAttributes(TypedDict, total=False):
         email: str
+        phone: str
+        user_updates: dict
 
-    def updateUser(self, user_id: str, attributes: UpdateUserAttributes) -> Union[PassageUserType, PassageError]:
+    def updateUser(self, user_id: str, attributes: UserAttributes) -> Union[PassageUserType, PassageError]:
         if self.passage_apikey == "":
             raise PassageError("No Passage API key provided.")
         
@@ -251,14 +341,10 @@ class Passage():
             raise PassageError("Could not delete user")
 
 
-    class UpdateUserAttributes(TypedDict, total=False):
-        email: str
-        phone: str
-
     """
     Create Passage User
     """
-    def createUser(self, userAttributes: UpdateUserAttributes) -> Union[PassageUserType, PassageError]:
+    def createUser(self, userAttributes: UserAttributes) -> Union[PassageUserType, PassageError]:
         if not ("phone" in userAttributes or "email" in userAttributes):
             raise PassageError("either phone or email must be provided to create the user")
 
@@ -280,11 +366,25 @@ class Passage():
         except Exception as e:
             raise PassageError("Could not create user")
 
+class PassageDevice:
+    def __init__(self, fields={}):
+        self.id = fields["id"]
+        try:
+            self.created_at = datetime.strptime(time_to_milliseconds(fields["created_at"]),"%Y-%m-%dT%H:%M:%S.%fZ")
+        except:
+            self.created_at = datetime.strptime(fields["created_at"],"%Y-%m-%dT%H:%M:%SZ")
+        try:
+            self.updated_at = datetime.strptime(time_to_milliseconds(fields["updated_at"]),"%Y-%m-%dT%H:%M:%S.%fZ")
+        except:
+            self.updated_at = datetime.strptime(fields["updated_at"],"%Y-%m-%dT%H:%M:%SZ")
+        self.friendly_name = fields["friendly_name"]
+        self.usage_count = fields["usage_count"]
+
 class PassageMagicLink:
     def __init__(self, magic_link_id, fields={}):
         self.id = magic_link_id
         self.secret = fields["secret"]
-        self.activated = fields["secret"]
+        self.activated = fields["activated"]
         self.user_id = fields["user_id"]
         self.app_id = fields["app_id"]
         self.identifier = fields["identifier"]
@@ -292,6 +392,7 @@ class PassageMagicLink:
         self.redirect_url = fields["redirect_url"]
         self.url = fields["url"]
         self.ttl = fields["ttl"]
+
 class PassageUser:
     def __init__(self, user_id, fields={}):
         self.id = user_id
@@ -299,6 +400,7 @@ class PassageUser:
         self.phone = fields["phone"]
         self.status = fields["status"]
         self.email_verified = fields["email_verified"]
+        self.user_metadata = fields["user_metadata"]
         try:
             self.created_at = datetime.strptime(time_to_milliseconds(fields["created_at"]),"%Y-%m-%dT%H:%M:%S.%fZ")
         except:
